# 词法分析科学计算器

> http://gogs.mkyr.fun:99/myuan/elang

## 目标

本节的目标是实现包含一些初等函数的计算器, 允许输入的例子如下:
```
# 允许以#开始写注释, #到行尾的将被忽略

1 + sin(degree(30)) # 角度制的30度
2 - cos(pi) / 10 # 默认为弧度制, 支持定义常量
3 * log(10, 100) # 以10为底, 取100的对数
4 / (2^3 / 4^2) # 当然仍然还有括号可以用
```

## 代码流程

上一节中实现功能时边读入边解析和计算, 如果按照这样的逻辑完成本节的目标, 代码将会冗长又固执, 难以阅读和修改. 本节的代码清晰地分为了两个部分: 

1. 词法分析, 将文本拆成不同的词单元, 这里靠正则表达式拆  
2. 语法分析, 根据不同的模式分析语法生成语法树, 上文中语法大致可以分为  
2.1 a + b 中缀表达式  
2.2 f(a, b, c) 函数调用式  
2.3 a * (b + c) / (f(a) + 2) 带括号的前二者
3. 解释执行语法树中的内容

## 词法分析

此处的「词」是一个泛指, 意为某些同类字的集合, 如「12334.235」是「数字和小数点」的集合是词, 「sin」是「字母」的集合是词, 「# 注释文本」也是词. 

这里定义了一些词类型
```
.版本 2

.常量 词类型枚举
.常量 词类_数字, "1"
.常量 词类_整数部分, "2"
.常量 词类_小数部分, "3"
.常量 词类_标识符, "4"
.常量 词类_算符, "5"
.常量 词类_括号, "6"
.常量 词类_注释, "7"
.常量 词类_空格, "8"
.常量 词类型枚举数量, "8"

```
词类_整数部分和词类_小数部分是为了迁就小数匹配, 词法分析的时候是跳过这两个的. 
词法分析的时候逻辑很简单, 伪代码大约是:
```
对于每一个词类
    如果 是整数部分 或 小数部分
        到循环尾
    end

    如果 正则匹配对当前剩余文本匹配成功
        记录匹配结果和类型
        文本游标往右走匹配结果那么长
        跳出循环
    end
end
```

这样一趟后就得到了这样的结果

```
> 词法分析 sin(x) + cos(pi * y) - ln(x) * 123.456 # 注释
用时 0ms
当前类型: 4, 内容: sin
当前类型: 6, 内容: (
当前类型: 4, 内容: x
当前类型: 6, 内容: )
当前类型: 5, 内容: +
当前类型: 4, 内容: cos
当前类型: 6, 内容: (
当前类型: 4, 内容: pi
当前类型: 5, 内容: *
当前类型: 4, 内容: y
当前类型: 6, 内容: )
当前类型: 5, 内容: -
当前类型: 4, 内容: ln
当前类型: 6, 内容: (
当前类型: 4, 内容: x
当前类型: 6, 内容: )
当前类型: 5, 内容: *
当前类型: 1, 内容: 123.456
当前类型: 7, 内容: # 注释
----
> 
```


## 语法树设计

语法树通常没有定型, 需因地制宜设计. 数学表达式通常可以视为一个纯的无求值顺序问题的语言. 可以做如下规约:
```
0 ± x -> ±(0, x)
x ± 0 -> ±(x, 0)
a + b -> +(a, b)
12345 -> +(12345, 0)
```
这样的话所有的计算表达式都可以写作 函(函甲, 函乙, 函丙, 函丁...)
> 函: 指函数

那么语法树的每一个节点应当包含如下信息

- 原始文本, debug用
- 函数名
- 参数数组, 类型为节点

但是很遗憾, 这样的代码会报递归定义错误, 我又没找到怎么把一个指针重新解释成某个自定义类型<sup>等我开始扩展易语言语法的时候首先要处理这个事情</sup>, 那只能自己申请内存管理数据了, 按C的写法数据结构定义如下: 

```C
{
    char     函数名[16],  // 定长16, 再长不支持了
    int32_t  参数量,      // 
    int32_t  参数数组容量, // 参数数组指针后有多大容量, 按照每次*2扩充
    uint64_t *参数数组指针, // 按长整数, 未来迁移到x64时兼容
}
// 此结构单个长度为 16+8+8 = 32 字节
```
内存结构如图: 
<svg id="SvgjsSvg1006" width="611" height="563" xmlns="http://www.w3.org/2000/svg" version="1.1" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:svgjs="http://svgjs.com/svgjs"><defs id="SvgjsDefs1007"><marker id="SvgjsMarker1026" markerWidth="14" markerHeight="10" refX="10" refY="5" viewBox="0 0 14 10" orient="auto" markerUnits="userSpaceOnUse" stroke-dasharray="0,0"><path id="SvgjsPath1027" d="M0,0 L14,5 L0,10 L0,0" fill="#323232" stroke="#323232" stroke-width="1"></path></marker><marker id="SvgjsMarker1038" markerWidth="14" markerHeight="10" refX="10" refY="5" viewBox="0 0 14 10" orient="auto" markerUnits="userSpaceOnUse" stroke-dasharray="0,0"><path id="SvgjsPath1039" d="M0,0 L14,5 L0,10 L0,0" fill="#323232" stroke="#323232" stroke-width="1"></path></marker><marker id="SvgjsMarker1072" markerWidth="14" markerHeight="10" refX="10" refY="5" viewBox="0 0 14 10" orient="auto" markerUnits="userSpaceOnUse" stroke-dasharray="0,0"><path id="SvgjsPath1073" d="M0,0 L14,5 L0,10 L0,0" fill="#323232" stroke="#323232" stroke-width="1"></path></marker><marker id="SvgjsMarker1100" markerWidth="14" markerHeight="10" refX="10" refY="5" viewBox="0 0 14 10" orient="auto" markerUnits="userSpaceOnUse" stroke-dasharray="0,0"><path id="SvgjsPath1101" d="M0,0 L14,5 L0,10 L0,0" fill="#323232" stroke="#323232" stroke-width="1"></path></marker><marker id="SvgjsMarker1136" markerWidth="14" markerHeight="10" refX="10" refY="5" viewBox="0 0 14 10" orient="auto" markerUnits="userSpaceOnUse" stroke-dasharray="0,0"><path id="SvgjsPath1137" d="M0,0 L14,5 L0,10 L0,0" fill="#323232" stroke="#323232" stroke-width="1"></path></marker></defs><g id="SvgjsG1008" transform="translate(25,136)"><path id="SvgjsPath1009" d="M 0 0L 160 0L 160 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1010"><text id="SvgjsText1011" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="140px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="2.375" transform="rotate(0)"><tspan id="SvgjsTspan1012" dy="16" x="80"><tspan id="SvgjsTspan1013" style="text-decoration:;">函数名</tspan></tspan><tspan id="SvgjsTspan1014" dy="16" x="80"><tspan id="SvgjsTspan1015" style="text-decoration:;">u8[16]</tspan></tspan></text></g></g><g id="SvgjsG1016" transform="translate(186,136)"><path id="SvgjsPath1017" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1018"><text id="SvgjsText1019" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1020" transform="translate(227,136)"><path id="SvgjsPath1021" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1022"><text id="SvgjsText1023" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1024"><path id="SvgjsPath1025" d="M206 135L206 100.5L206 100.5L206 68.6" stroke="#323232" stroke-width="2" fill="none" marker-end="url(#SvgjsMarker1026)"></path></g><g id="SvgjsG1028" transform="translate(146,25)"><path id="SvgjsPath1029" d="M 0 0L 120 0L 120 40L 0 40Z" stroke="none" fill="none"></path><g id="SvgjsG1030"><text id="SvgjsText1031" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="120px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="2.375" transform="rotate(0)"><tspan id="SvgjsTspan1032" dy="16" x="60"><tspan id="SvgjsTspan1033" style="text-decoration:;">当前参数数量</tspan></tspan><tspan id="SvgjsTspan1034" dy="16" x="60"><tspan id="SvgjsTspan1035" style="text-decoration:;">int32</tspan></tspan></text></g></g><g id="SvgjsG1036"><path id="SvgjsPath1037" d="M247 177L247 204L247 204L247 228.39999999999998" stroke="#323232" stroke-width="2" fill="none" marker-end="url(#SvgjsMarker1038)"></path></g><g id="SvgjsG1040" transform="translate(185,232)"><path id="SvgjsPath1041" d="M 0 0L 120 0L 120 40L 0 40Z" stroke="none" fill="none"></path><g id="SvgjsG1042"><text id="SvgjsText1043" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="120px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="2.375" transform="rotate(0)"><tspan id="SvgjsTspan1044" dy="16" x="60"><tspan id="SvgjsTspan1045" style="text-decoration:;">当前参数数组容量</tspan></tspan><tspan id="SvgjsTspan1046" dy="16" x="60"><tspan id="SvgjsTspan1047" style="text-decoration:;">int32</tspan></tspan></text></g></g><g id="SvgjsG1048" transform="translate(265,136)"><path id="SvgjsPath1049" d="M 0 0L 80 0L 80 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1050"><text id="SvgjsText1051" font-family="微软雅黑" text-anchor="middle" font-size="12px" width="60px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="12px" weight="400" font-style="" opacity="1" y="3.5" transform="rotate(0)"><tspan id="SvgjsTspan1052" dy="15" x="40"><tspan id="SvgjsTspan1053" style="text-decoration:;">数组头指针</tspan></tspan><tspan id="SvgjsTspan1054" dy="15" x="40"><tspan id="SvgjsTspan1055" style="text-decoration:;">uint64</tspan></tspan></text></g></g><g id="SvgjsG1056" transform="translate(106,341)"><path id="SvgjsPath1057" d="M 0 0L 480 0L 480 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1058"><text id="SvgjsText1059" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="460px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"><tspan id="SvgjsTspan1060" dy="16" x="240"><tspan id="SvgjsTspan1061" style="text-decoration:;">3, 4, 5, 6, 7, 8...</tspan></tspan></text></g></g><g id="SvgjsG1062" transform="translate(25,341)"><path id="SvgjsPath1063" d="M 0 0L 80 0L 80 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1064"><text id="SvgjsText1065" font-family="微软雅黑" text-anchor="middle" font-size="12px" width="60px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="12px" weight="400" font-style="" opacity="1" y="3.5" transform="rotate(0)"><tspan id="SvgjsTspan1066" dy="15" x="40"><tspan id="SvgjsTspan1067" style="text-decoration:;">数组元素1</tspan></tspan><tspan id="SvgjsTspan1068" dy="15" x="40"><tspan id="SvgjsTspan1069" style="text-decoration:;">uint64</tspan></tspan></text></g></g><g id="SvgjsG1070"><path id="SvgjsPath1071" d="M305 177L305 279.5L65 279.5L65 337.4" stroke="#323232" stroke-width="2" fill="none" marker-end="url(#SvgjsMarker1072)"></path></g><g id="SvgjsG1074" transform="translate(243,498)"><path id="SvgjsPath1075" d="M 0 0L 160 0L 160 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1076"><text id="SvgjsText1077" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="140px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="2.375" transform="rotate(0)"><tspan id="SvgjsTspan1078" dy="16" x="80"><tspan id="SvgjsTspan1079" style="text-decoration:;">函数名</tspan></tspan><tspan id="SvgjsTspan1080" dy="16" x="80"><tspan id="SvgjsTspan1081" style="text-decoration:;">u8[16]</tspan></tspan></text></g></g><g id="SvgjsG1082" transform="translate(404,498)"><path id="SvgjsPath1083" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1084"><text id="SvgjsText1085" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1086" transform="translate(445,498)"><path id="SvgjsPath1087" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1088"><text id="SvgjsText1089" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1090" transform="translate(483,498)"><path id="SvgjsPath1091" d="M 0 0L 80 0L 80 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1092"><text id="SvgjsText1093" font-family="微软雅黑" text-anchor="middle" font-size="12px" width="60px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="12px" weight="400" font-style="" opacity="1" y="3.5" transform="rotate(0)"><tspan id="SvgjsTspan1094" dy="15" x="40"><tspan id="SvgjsTspan1095" style="text-decoration:;">数组头指针</tspan></tspan><tspan id="SvgjsTspan1096" dy="15" x="40"><tspan id="SvgjsTspan1097" style="text-decoration:;">uint64</tspan></tspan></text></g></g><g id="SvgjsG1098"><path id="SvgjsPath1099" d="M65 382L65 518L237.39999999999998 518" stroke="#323232" stroke-width="2" fill="none" marker-end="url(#SvgjsMarker1100)"></path></g><g id="SvgjsG1102" transform="translate(105,341)"><path id="SvgjsPath1103" d="M 0 0L 80 0L 80 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1104"><text id="SvgjsText1105" font-family="微软雅黑" text-anchor="middle" font-size="12px" width="60px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="12px" weight="400" font-style="" opacity="1" y="3.5" transform="rotate(0)"><tspan id="SvgjsTspan1106" dy="15" x="40"><tspan id="SvgjsTspan1107" style="text-decoration:;">数组元素2</tspan></tspan><tspan id="SvgjsTspan1108" dy="15" x="40"><tspan id="SvgjsTspan1109" style="text-decoration:;">uint64</tspan></tspan></text></g></g><g id="SvgjsG1110" transform="translate(243,439)"><path id="SvgjsPath1111" d="M 0 0L 160 0L 160 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1112"><text id="SvgjsText1113" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="140px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="2.375" transform="rotate(0)"><tspan id="SvgjsTspan1114" dy="16" x="80"><tspan id="SvgjsTspan1115" style="text-decoration:;">函数名</tspan></tspan><tspan id="SvgjsTspan1116" dy="16" x="80"><tspan id="SvgjsTspan1117" style="text-decoration:;">u8[16]</tspan></tspan></text></g></g><g id="SvgjsG1118" transform="translate(404,439)"><path id="SvgjsPath1119" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1120"><text id="SvgjsText1121" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1122" transform="translate(445,439)"><path id="SvgjsPath1123" d="M 0 0L 40 0L 40 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1124"><text id="SvgjsText1125" font-family="微软雅黑" text-anchor="middle" font-size="13px" width="20px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="13px" weight="400" font-style="" opacity="1" y="10.375" transform="rotate(0)"></text></g></g><g id="SvgjsG1126" transform="translate(483,439)"><path id="SvgjsPath1127" d="M 0 0L 80 0L 80 40L 0 40Z" stroke="rgba(50,50,50,1)" stroke-width="2" fill-opacity="1" fill="#ffffff"></path><g id="SvgjsG1128"><text id="SvgjsText1129" font-family="微软雅黑" text-anchor="middle" font-size="12px" width="60px" fill="#323232" font-weight="400" align="middle" lineHeight="125%" anchor="middle" family="微软雅黑" size="12px" weight="400" font-style="" opacity="1" y="3.5" transform="rotate(0)"><tspan id="SvgjsTspan1130" dy="15" x="40"><tspan id="SvgjsTspan1131" style="text-decoration:;">数组头指针</tspan></tspan><tspan id="SvgjsTspan1132" dy="15" x="40"><tspan id="SvgjsTspan1133" style="text-decoration:;">uint64</tspan></tspan></text></g></g><g id="SvgjsG1134"><path id="SvgjsPath1135" d="M145 382L145 459L239.39999999999998 459" stroke="#323232" stroke-width="2" fill="none" marker-end="url(#SvgjsMarker1136)"></path></g></svg>

终于可以开始写递归下降器了. 

## 其他

细微之处的东西在代码的注释里. 

本节代码使用 git 组织, 每一个里程碑我都会创建一个 commit, 你可以通过 git 回溯来分步查看我的编码过程. 如果你不会 git, 可以下载一个 github desktop, 可以直接在 GUI 上看到各个 commit 的内容. 

### 额外阅读
- 正则表达式 
- - https://regex101.com/ 
- - https://www.runoob.com/regexp/regexp-syntax.html
- 递归下降器 
- - https://zhuanlan.zhihu.com/p/31271879
- - https://zh.wikipedia.org/wiki/%E9%80%92%E5%BD%92%E4%B8%8B%E9%99%8D%E8%A7%A3%E6%9E%90%E5%99%A8

## 下节预告

本节代码仍然使用纯的易语言做文本分析, 但是如果你详细读了这份代码, 会发现大量的重复, 之后我将使用 JavaScript 的 [peggyjs](https://peggyjs.org/online) 库来重新做这件事, 同时, 如果你认真读了本节代码, 你也可以轻松读懂 peggyjs 的大部分东西. 

> *不过严格来说, 正则表达式也是 DSL, 因此用了正则表达式就不算是纯易语言了.*
